#include <ros/ros.h>
#include <geometry_msgs/Twist.h>
#include <nav_msgs/Odometry.h>
#include <std_msgs/Float64.h>
#include <scamper_move/ScamperMoveAction.h>
#include <actionlib/server/simple_action_server.h>
#include <cmath>

using namespace scamper_move;
using actionlib::SimpleActionServer;
typedef SimpleActionServer<ScamperMoveAction> Server;

geometry_msgs::Pose g_pose;
double g_current_dist = .0;

/*** Actionを実行したときに呼び出されるコールバック関数 ***/
void actMoveDirection(const ScamperMoveGoalConstPtr &goal, 
      Server *server, ros::Publisher *pub_vel, double max_vel)
{
  static double old_dist = .0;
  geometry_msgs::Twist stop_vel, target_vel;
  double goal_x = goal->target_pt.x, goal_y = goal->target_pt.y;
  /*** 目標地点の方向からVx, Vyを決定 ***/
  if(goal_x == 0 && goal_y == 0){
    target_vel = stop_vel;
  } else{
    double target_dir = atan2(goal_y, goal_x);
    target_vel.linear.x = max_vel * cos(target_dir);
    target_vel.linear.y = max_vel * sin(target_dir);
  }
  /*** 目標地点までの距離を算出 ***/
  double target_dist = sqrt(goal_x * goal_x + goal_y * goal_y);
  old_dist = g_current_dist;
  /*** ループ周期の設定 ***/
  ros::Rate loop_rate(10);
  /*** feedbackとresultの宣言 ***/
  ScamperMoveFeedback feedback;
  ScamperMoveResult result;
  /*** 目標地点までの移動制御 ***/
  while(ros::ok() && !server->isPreemptRequested()){
    /*** 目標地点まで到達した場合 ***/
    if((g_current_dist - old_dist) >= target_dist){
      /*** 結果のメッセージを設定 ***/
      result.msg = "Succeeded";
      break;
    }
    /*** 目標地点まで到達していない場合 ***/
    else{
      /*** 現在の座標を送信 ***/
      feedback.pose = g_pose;
      server->publishFeedback(feedback);
      pub_vel->publish(target_vel);
    }
    loop_rate.sleep();
  }
  pub_vel->publish(stop_vel);
  if(server->isPreemptRequested()) server->setPreempted();
  else server->setSucceeded(result);
}

/*** robot_poseを受信したときに呼ばれるコールバック関数 ***/
void onPoseSubscribed(const nav_msgs::Odometry &pose)
{
  g_pose = pose.pose.pose;
}

/*** distanceを受信したときに呼ばれるコールバック関数 ***/
void onDistSubscribed(const std_msgs::Float64 &dist)
{
  g_current_dist = dist.data;
}

int main(int argc, char** argv)
{
  /*** ROSの初期化 ***/
  ros::init(argc, argv, "omni_move");
  ros::NodeHandle nh;
  /*** Parameterの読み込み ***/
  double max_vel = nh.param<double>("max_vel", 0.1);
  /*** Publisherの初期化 ***/
  ros::Publisher pub_vel = nh.advertise<geometry_msgs::Twist>("/scamper_driver/cmd_robot_vel", 10);
  /*** Subscriberの初期化 ***/
  ros::Subscriber sub_odo = nh.subscribe("/scamper_driver/robot_pose", 10, onPoseSubscribed);
  ros::Subscriber sub_dist = nh.subscribe("/scamper_driver/distance", 10, onDistSubscribed);
  /*** Action serverの初期化 ***/
  Server server(nh, "omni_move", boost::bind(&actMoveDirection, _1, 
                &server, &pub_vel, max_vel), false);
  /*** Action serverの開始 ***/
  server.start();
  /*** コールバックを待機 ***/
  ros::spin();
  return 0;
}