#include <ros/ros.h>
#include <std_msgs/Float32MultiArray.h>
#include <scamper_move/ScamperMoveAction.h>
#include <actionlib/client/simple_action_client.h>

using namespace scamper_move;
using actionlib::SimpleActionClient;
using actionlib::SimpleClientGoalState;
typedef SimpleActionClient<ScamperMoveAction> Client;

std::vector<float> g_sonar;

void onSonarSubscribed(const std_msgs::Float32MultiArray &sonar)
{
  g_sonar = sonar.data;
}

int main(int argc, char** argv)
{
  /*** ROSノードの初期化 ***/
  ros::init(argc, argv, "wander");
  ros::NodeHandle nh;
  /*** Parameterの読み込み ***/
  std::string robot_type;
  double obst_dist = nh.param<double>("obst_dist", 0.2);
  nh.getParam("/scamper_driver/robot_type", robot_type);
  /*** Subscriberの設定 ***/
  ros::Subscriber sub_sonar = nh.subscribe("/scamper_sonar/sonar_dist", 10, onSonarSubscribed);
  /*** actionlib clientの初期化 ***/
  Client client("omni_move", true);
  client.waitForServer();
  /*** 目標地点のリストを生成 ***/
  std::vector<std::array<float, 2>> target_pts;
  if(robot_type == "Omni3WD"){
    target_pts = std::vector<std::array<float, 2>>
    {
      std::array<float, 2>{0.25, -0.43}, std::array<float, 2>{0.25, 0.43},
      std::array<float, 2>{-0.5, .0}
    };
  } else if(robot_type == "Mecanum4WD"){
    target_pts = std::vector<std::array<float, 2>>
    {
      std::array<float, 2>{.0, -0.5}, std::array<float, 2>{0.5, .0},
      std::array<float, 2>{.0, 0.5}, std::array<float, 2>{-0.5, .0}
    };
  } else{
    ros::shutdown();
  }
  std::array<bool, 4> is_obst = {false, false, false, false};
  bool is_stop = false;
  /*** コールバックを別スレッドで待機 ***/
  ros::AsyncSpinner spinner(1);
  spinner.start();
  /*** ループ周期の設定 ***/
  ros::Rate loop_rate(10);
  /*** Scamperをランダム（3または4方向）に移動 ***/
  ScamperMoveGoal goal;
  while(ros::ok()){
    /*** 現在の時間から移動する方向を決定 ***/
    unsigned long t = ros::Time::now().toNSec();
    /*** 進行方向の設定 ***/
    unsigned int dir = t % target_pts.size();
    is_stop = false;
    /*** 障害物のチェック ***/
    for(int i = 0; i < g_sonar.size(); i++){
      if(g_sonar[i] < obst_dist && g_sonar[i] != .0f) is_obst[i] = true;
      else is_obst[i] = false;
    }
    /*** 障害物のある方向には進まないように設定 ***/
    if(is_obst[dir]){
      for(int i = 0; i < target_pts.size() - 1; i++){
        if(++dir == target_pts.size()) dir = 0;
        if(!is_obst[dir]) break;
        if(i == target_pts.size() - 2) is_stop = true;
      }
    }
    /*** 目標地点の設定 ***/
    if(is_stop){
      loop_rate.sleep();
      continue;
    } else{
      goal.target_pt.x = target_pts[dir][0];
      goal.target_pt.y = target_pts[dir][1];
    }
    /*** Actionの呼び出し ***/
    client.sendGoal(goal);
    /*** Actionが終了するまで待機 ***/
    while(ros::ok()){
      /*** 障害物のチェック ***/
      for(int i = 0; i < g_sonar.size(); i++){
        if(g_sonar[i] < obst_dist && g_sonar[i] != .0f) is_obst[i] = true;
        else is_obst[i] = false;
      }
      /*** 進行方向に障害物が存在すればActionを中断 ***/
      if(is_obst[dir]) {
        client.cancelGoal();
        ROS_INFO("Obstacle was detected");
        while(client.getState() != SimpleClientGoalState::PREEMPTED);
        break;
      }
      if(client.getState() == SimpleClientGoalState::SUCCEEDED) break;
      loop_rate.sleep();
    }
  }
  return 0;
}