#include <ros/ros.h>
#include <message_filters/subscriber.h>
#include <message_filters/sync_policies/approximate_time.h>
#include <sensor_msgs/Image.h>
#include <sensor_msgs/CameraInfo.h>
#include <cv_bridge/cv_bridge.h>
#include <geometry_msgs/Twist.h>
#include <nodelet/nodelet.h>
#include <pluginlib/class_list_macros.h>
#include <opencv2/opencv.hpp>
#include <thread>

using namespace sensor_msgs;
using namespace geometry_msgs;
using namespace cv_bridge;
using message_filters::sync_policies::ApproximateTime;
using message_filters::Synchronizer;

typedef ApproximateTime<Image, Image, CameraInfo, CameraInfo> SyncPolicy;

namespace scamper_vision
{

class FollowColor : public nodelet::Nodelet
{
public:
  FollowColor();
  ~FollowColor();
  virtual void onInit();
private:
  void extractColor(const cv::Mat &src, cv::Mat &dst);
  void labeling(const cv::Mat &bin, cv::Rect &rect);
  void calcDistance(const cv::Mat &img_l, const cv::Mat &img_r,
                    const cv::Rect &rect_l, const cv::Rect &rect_r,
                    double f, double baseline, cv::Point3d &pt_3d);
  Twist calcVelocity(cv::Point3d &pt_3d);
  void onImgsSubscribed(const ImageConstPtr &img_l, 
                        const ImageConstPtr &img_r,
                        const CameraInfoConstPtr &info_l, 
                        const CameraInfoConstPtr &info_r);
  ros::NodeHandle nh_;
  message_filters::Subscriber<Image> *sub_img_l_, *sub_img_r_;
  message_filters::Subscriber<CameraInfo> *sub_info_l_, *sub_info_r_;
  Synchronizer<SyncPolicy> *sync_msgs_;
  ros::Publisher pub_vel_;
  /*** Parameters ***/
  int hue_min_, hue_max_, sat_min_, sat_max_, val_min_, val_max_;
  int sz_min_, find_area_;
  double x_max_, x_target_, x_deadzone_;
  double r_max_, r_deadzone_;
  double vel_x_max_, vel_r_max_;
  double score_min_;
};

FollowColor::FollowColor()
{

}

FollowColor::~FollowColor()
{

}

/*** モジュールが読み込まれると自動的に呼び出される関数 ***/
void FollowColor::onInit()
{
  using message_filters::Subscriber;
  /*** ノードハンドラの初期化 ***/
  nh_ = this->getNodeHandle();
  /*** Parameterの読み込み ***/
  hue_min_ = nh_.param<int>("hue/min", 50);
  hue_max_ = nh_.param<int>("hue/max", 65);
  sat_min_ = nh_.param<int>("sat/min", 100);
  sat_max_ = nh_.param<int>("sat/max", 255);
  val_min_ = nh_.param<int>("val/min", 50);
  val_max_ = nh_.param<int>("val/max", 255);
  sz_min_ = nh_.param<int>("size/min", 400);
  find_area_ = nh_.param<int>("find_area", 10);
  x_max_ = nh_.param<double>("distance/max", 1.0);
  x_target_ = nh_.param<double>("distance/target", 0.3);
  x_deadzone_ = nh_.param<double>("distance/dead_zone", 0.05);
  r_max_ = nh_.param<double>("rotation/max", 25.0);
  r_deadzone_ = nh_.param<double>("rotation/dead_zone", 5.0);
  vel_x_max_ = nh_.param<double>("vel/x/max", 0.20);
  vel_r_max_ = nh_.param<double>("vel/r/max", 50.0);
  score_min_ = nh_.param<double>("score/min", 0.70);
  /*** DegreeからRadianへの変換 ***/
  r_max_ *= (M_PI / 180.0);
  r_deadzone_ *= (M_PI / 180.0);
  vel_r_max_ *= (M_PI / 180.0);
  /*** message_filtersの初期化 ***/
  sub_img_l_ = new Subscriber<Image>(nh_, "/scamper_eyes/img/left", 10);
  sub_img_r_ = new Subscriber<Image>(nh_, "/scamper_eyes/img/right", 10);
  sub_info_l_ = new Subscriber<CameraInfo>(nh_, "/scamper_eyes/info/left", 10);
  sub_info_r_ = new Subscriber<CameraInfo>(nh_, "/scamper_eyes/info/right", 10);
  sync_msgs_ = new Synchronizer<SyncPolicy>(SyncPolicy(10), *sub_img_l_, *sub_img_r_, *sub_info_l_, *sub_info_r_);
  sync_msgs_->registerCallback(boost::bind(&FollowColor::onImgsSubscribed, this, _1, _2, _3, _4));
  /*** Publisherの初期化 ***/
  pub_vel_ = nh_.advertise<Twist>("/scamper_driver/cmd_robot_vel", 10);
}

/*** 特定色の領域を抽出 ***/
void FollowColor::extractColor(const cv::Mat &src, cv::Mat &dst)
{
  /*** srcをHSV表色系に変換 ***/
  cv::Mat hsv;
  cv::cvtColor(src, hsv, CV_BGR2HSV);
  /*** 出力画像を0で初期化 ***/
  dst = cv::Mat(src.size(), CV_8UC1, cv::Scalar(0));
  /*** H・S・Vがしきい値内の画素のみ255を代入 ***/
  for(int y = 0; y < hsv.rows; y++){
    const uchar *ptr_hsv = hsv.ptr<uchar>(y);
    uchar *ptr_dst = dst.ptr<uchar>(y);
    for(int x = 0; x < hsv.cols; x++){
      uchar hue = ptr_hsv[3 * x + 0];
      uchar sat = ptr_hsv[3 * x + 1];
      uchar val = ptr_hsv[3 * x + 2];
      if(hue < hue_min_ || hue > hue_max_) continue;
      if(sat < sat_min_ || sat > sat_max_) continue;
      if(val < val_min_ || val > val_max_) continue;
      ptr_dst[x] = 255;
    }
  }
}

/*** ラベリング処理 ***/
void FollowColor::labeling(const cv::Mat &bin, cv::Rect &rect)
{
  cv::Mat label_img, stats, centroids;
  int n = cv::connectedComponentsWithStats(bin, label_img, stats, centroids, 8);
  /*** ラベリング領域が最大のものを選択 ***/
  int sz = 0;
  for(int i = 1; i < n; i++){
    int *stat = stats.ptr<int>(i);
    int tmp_sz = stat[cv::ConnectedComponentsTypes::CC_STAT_AREA];
    if(sz < tmp_sz){
      sz = tmp_sz;
      rect.x = stat[cv::ConnectedComponentsTypes::CC_STAT_LEFT];
      rect.y = stat[cv::ConnectedComponentsTypes::CC_STAT_TOP];
      rect.width = stat[cv::ConnectedComponentsTypes::CC_STAT_WIDTH];
      rect.height = stat[cv::ConnectedComponentsTypes::CC_STAT_HEIGHT];
    }
  }
}

void FollowColor::calcDistance(const cv::Mat &img_l,
                               const cv::Mat &img_r,
                               const cv::Rect &rect_l,
                               const cv::Rect &rect_r,
                               double f, double baseline,
                               cv::Point3d &pt_3d)
{
  pt_3d = cv::Point3d();
  int cx = img_l.cols / 2, cy = img_l.rows / 2;
  if(rect_l.width * rect_l.height < sz_min_ || rect_r.width * rect_r.height < sz_min_) return;
  /*** 探索用のテンプレート画像を作成 ***/
  cv::Mat temp;
  cv::cvtColor(img_l(rect_l), temp, CV_BGR2GRAY);
  /*** 探索用の入力画像を作成 ***/
  cv::Rect find_rect(rect_r.x - find_area_, rect_l.y,
                     rect_l.width + find_area_ * 2 + 1, rect_l.height);
  if(find_rect.x < 0){
    find_rect.x = 0;
  }
  if(find_rect.x + find_rect.width > img_r.cols){
    find_rect.width = img_r.cols - find_rect.x;
  }
  cv::Mat input;
  cv::cvtColor(img_r(find_rect), input, CV_BGR2GRAY);
  /*** テンプレート画像に対応する領域を入力画像（右画像）から探索 ***/
  cv::Mat result(1, input.cols - temp.cols + 1, CV_32FC1);
  cv::matchTemplate(input, temp, result, CV_TM_CCORR_NORMED);
  /*** テンプレートマッチングの結果が最大になる点を探索 ***/
  double min, max;
  cv::Point min_loc, max_loc;
  cv::minMaxLoc(result, &min, &max, &min_loc, &max_loc);
  /*** マッチスコアがしきい値以下なら関数を抜ける ***/
  if(max < score_min_) return;
  /*** 物体の３次元座標を計算 ***/
  int x1 = rect_l.x + rect_l.width / 2 - cx;
  int x2 = rect_r.x + max_loc.x + rect_l.width / 2 - cx;
  int y1 = rect_l.y + rect_l.height / 2 - cy;
  if(x1 == x2){
    pt_3d.x = std::numeric_limits<double>().max();
    pt_3d.y = pt_3d.z = 0;
  } else{
    pt_3d.x = baseline * f / (x1 - x2);
    pt_3d.y = -pt_3d.x * x1 / f + baseline / 2;
    pt_3d.z = -pt_3d.x * y1 / f;
  }
}

/*** 物体の位置情報からロボットの速度を計算 ***/
Twist FollowColor::calcVelocity(cv::Point3d &pt_3d)
{
  Twist vel;
  if(pt_3d.z <= 0) return vel;
  double dx = pt_3d.x - x_target_;  // 目標距離との差
  double obj_dir = atan2(pt_3d.y, pt_3d.x);  // 物体の方向
  /*** 並進方向の速度（x方向のみ）を計算 ***/
  if(fabs(dx) > x_deadzone_){
    vel.linear.x = dx * vel_x_max_ / x_max_;
    if(vel.linear.x > vel_x_max_) vel.linear.x = vel_x_max_;
  }
  /*** 回転方向の速度を計算 ***/
  if(fabs(obj_dir) > r_deadzone_){
    vel.angular.z = obj_dir * vel_r_max_ / r_max_;
    if(vel.angular.z > vel_r_max_) vel.angular.z = vel_r_max_;
    else if(vel.angular.z < -vel_r_max_) vel.angular.z = -vel_r_max_;
  }
  return vel;
}

/*** 画像を受信すると呼ばれるコールバック関数 ***/
void FollowColor::onImgsSubscribed(const ImageConstPtr &img_l,
                                   const ImageConstPtr &img_r,
                                   const CameraInfoConstPtr &info_l,
                                   const CameraInfoConstPtr &info_r)
{
  /*** 変数の宣言 ***/
  std::array<CvImageConstPtr, 2> cv_img;
  std::array<cv::Mat, 2> bin;
  std::array<cv::Rect, 2> target_rect;
  /*** 受信した画像を変換 ***/
  cv_img[0] = toCvShare(img_l, img_l->encoding);
  cv_img[1] = toCvShare(img_r, img_r->encoding);
  if(cv_img[0]->image.empty() || cv_img[1]->image.empty()) return;
  /*** 特定の色の抽出〜ラベリング処理 ***/
  std::thread t1([&]{
    extractColor(cv_img[0]->image, bin[0]);
    labeling(bin[0], target_rect[0]);
  });
  std::thread t2([&]{
    extractColor(cv_img[1]->image, bin[1]);
    labeling(bin[1], target_rect[1]);
  });
  t1.join();
  t2.join();
  /*** 物体の3次元情報を算出 ***/
  cv::Point3d target_pt;
  double f = info_l->P[0];
  double baseline = fabs(info_r->P[3] / f);
  calcDistance(cv_img[0]->image, cv_img[1]->image, target_rect[0], target_rect[1],
               f, baseline, target_pt);
  /*** 物体の3次元情報を表示 ***/
  ROS_INFO("(%.2f, %.2f, %.2f)", target_pt.x, target_pt.y, target_pt.z);
  /*** ロボットの速度を計算してPublish ***/
  pub_vel_.publish(calcVelocity(target_pt));
}

}

PLUGINLIB_EXPORT_CLASS(scamper_vision::FollowColor, nodelet::Nodelet)